package com.github.polpetta.mezzotre.i18n;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.nio.charset.StandardCharsets;
import java.util.Enumeration;
import java.util.Locale;
import java.util.ResourceBundle;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.tools.ToolContext;
import org.apache.velocity.tools.ToolManager;
import org.apache.velocity.util.StringBuilderWriter;
import org.jetbrains.annotations.NotNull;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.parallel.Execution;
import org.junit.jupiter.api.parallel.ExecutionMode;
import org.mockito.ArgumentCaptor;

@Execution(ExecutionMode.CONCURRENT)
class TemplateContentGeneratorTest {

  private LocalizedMessageFactory fakeLocalizedMessageFactory;
  private TemplateContentGenerator templateContentGenerator;

  @BeforeEach
  void setUp() {
    fakeLocalizedMessageFactory = mock(LocalizedMessageFactory.class);
    templateContentGenerator = new TemplateContentGenerator(fakeLocalizedMessageFactory);
  }

  @Test
  void shouldCallMergeTemplateWithRightArguments() {
    final VelocityEngine fakeVelocityEngine = mock(VelocityEngine.class);
    final ToolManager fakeToolManager = mock(ToolManager.class);
    when(fakeToolManager.getVelocityEngine()).thenReturn(fakeVelocityEngine);
    final ToolContext fakeToolContext = mock(ToolContext.class);
    when(fakeToolManager.createContext()).thenReturn(fakeToolContext);
    when(fakeLocalizedMessageFactory.createVelocityToolManager(Locale.forLanguageTag("en-US")))
        .thenReturn(fakeToolManager);

    templateContentGenerator.mergeTemplate(
        ctx -> {
          ctx.put("testtesttest", "value");
        },
        "en-US",
        "a/fake/path.vm");

    final ArgumentCaptor<VelocityContext> velocityContextArgumentCaptor =
        ArgumentCaptor.forClass(VelocityContext.class);
    verify(fakeVelocityEngine, times(1))
        .mergeTemplate(
            eq("a/fake/path.vm"),
            eq(StandardCharsets.UTF_8.name()),
            velocityContextArgumentCaptor.capture(),
            any(StringBuilderWriter.class));

    final VelocityContext gotVelocityContext = velocityContextArgumentCaptor.getValue();
    assertEquals("value", gotVelocityContext.get("testtesttest"));
  }

  @Test
  void shouldCallLocaleFactoryWhenRetrievingKey() {
    final Locale enUS = Locale.forLanguageTag("en-US");

    class StubRB extends ResourceBundle {

      @Override
      protected Object handleGetObject(@NotNull String s) {
        return "testtest123";
      }

      @NotNull
      @Override
      public Enumeration<String> getKeys() {
        return new Enumeration<String>() {

          private boolean next = true;

          @Override
          public boolean hasMoreElements() {
            final boolean toRet = next;
            next = !next;
            return toRet;
          }

          @Override
          public String nextElement() {
            return "doens't matter";
          }
        };
      }
    }

    final StubRB stubRB = new StubRB();
    when(fakeLocalizedMessageFactory.createResourceBundle(any())).thenReturn(stubRB);

    final String got = templateContentGenerator.getString(enUS, "a.string");
    assertEquals("testtest123", got);
  }
}
