package com.github.polpetta.mezzotre.telegram.command;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

import com.github.polpetta.mezzotre.i18n.TemplateContentGenerator;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.telegram.CampaignUtil;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.request.BaseRequest;
import java.util.Collections;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.stream.Stream;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Timeout;
import org.junit.jupiter.api.parallel.Execution;
import org.junit.jupiter.api.parallel.ExecutionMode;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.ArgumentCaptor;

@Execution(ExecutionMode.CONCURRENT)
class CreateCampaignTest {

  private static Gson gson;
  private TemplateContentGenerator fakeTemplateContentGenerator;
  private UUIDGenerator fakeUUIDGenerator;
  private ChatUtil fakeChatUtil;
  private CampaignUtil fakeCampaignUtil;
  private CreateCampaign createCampaign;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
  }

  @BeforeEach
  void setUp() {

    fakeTemplateContentGenerator = mock(TemplateContentGenerator.class);
    fakeUUIDGenerator = mock(UUIDGenerator.class);
    fakeChatUtil = mock(ChatUtil.class);
    fakeCampaignUtil = mock(CampaignUtil.class);

    createCampaign =
        new CreateCampaign(
            fakeTemplateContentGenerator,
            Executors.newSingleThreadExecutor(),
            fakeUUIDGenerator,
            fakeChatUtil,
            fakeCampaignUtil);
  }

  public static Stream<Arguments> getMultipleValidCampaignEntries() {
    return Stream.of(
        Arguments.of("a simple name", "a simple description"),
        Arguments.of("a simple name", "a simple\nmultiline\n description"),
        Arguments.of(
            "an \uD83E\uDD21 emoji \uD83D\uDCAF name \uD83D\uDD25",
            "a very \uD83E\uDDE8 campaign description \uD83E\uDD16"),
        Arguments.of("一个简单的名字", "一个简单的描述"));
  }

  @ParameterizedTest
  @MethodSource("getMultipleValidCampaignEntries")
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  void shouldGenerateANewCampaign(String name, String description) throws Exception {

    final TgChat fakeTgChat = mock(TgChat.class);
    final ChatContext initialChatContext = new ChatContext();
    initialChatContext.setStep(0); // implicit but better specify just for clarity
    when(fakeTgChat.getChatContext()).thenReturn(initialChatContext);
    when(fakeTgChat.getLocale()).thenReturn("en-US");

    {
      when(fakeTemplateContentGenerator.mergeTemplate(
              any(), eq("en-US"), eq("template/telegram/createCampaign.0.vm")))
          .thenReturn("a string");
      final Update firstMessage =
          gson.fromJson(
              "{\n"
                  + "\"update_id\":10000,\n"
                  + "\"message\":{\n"
                  + "  \"date\":1441645532,\n"
                  + "  \"chat\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"type\": \"private\",\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"message_id\":1365,\n"
                  + "  \"from\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"text\":\"/createCampaign\"\n"
                  + "}\n"
                  + "}",
              Update.class);

      final CompletableFuture<Optional<BaseRequest<?, ?>>> got =
          createCampaign.process(fakeTgChat, firstMessage);
      final Optional<BaseRequest<?, ?>> baseRequestOptional = got.get();
      final BaseRequest<?, ?> gotResponse = baseRequestOptional.get();
      assertEquals("a string", gotResponse.getParameters().get("text"));

      verify(fakeChatUtil, times(1))
          .updateChatContext(fakeTgChat, "/createCampaign", 1, Collections.emptyMap());
    }

    {
      when(fakeTemplateContentGenerator.mergeTemplate(
              any(), eq("en-US"), eq("template/telegram/createCampaign.2.vm")))
          .thenReturn("a second string");
      final ChatContext chatContext = new ChatContext();
      chatContext.setStep(1);
      when(fakeTgChat.getChatContext()).thenReturn(chatContext);

      final Update secondMessageName =
          gson.fromJson(
              "{\n"
                  + "\"update_id\":10000,\n"
                  + "\"message\":{\n"
                  + "  \"date\":1441645532,\n"
                  + "  \"chat\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"type\": \"private\",\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"message_id\":1365,\n"
                  + "  \"from\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"text\":\""
                  + name
                  + "\"\n"
                  + "}\n"
                  + "}",
              Update.class);

      final CompletableFuture<Optional<BaseRequest<?, ?>>> got =
          createCampaign.process(fakeTgChat, secondMessageName);
      final Optional<BaseRequest<?, ?>> baseRequestOptional = got.get();
      final BaseRequest<?, ?> gotResponse = baseRequestOptional.get();
      assertEquals("a second string", gotResponse.getParameters().get("text"));

      final ArgumentCaptor<Map<String, Object>> capturedChatCtx =
          ArgumentCaptor.forClass(Map.class);
      verify(fakeChatUtil, times(1))
          .updateChatContext(
              eq(fakeTgChat), eq("/createCampaign"), eq(2), capturedChatCtx.capture());

      final Map<String, Object> chatCtxValue = capturedChatCtx.getValue();
      assertEquals(name, chatCtxValue.get("campaign_name"));
    }

    {
      when(fakeTemplateContentGenerator.mergeTemplate(
              any(), eq("en-US"), eq("template/telegram/createCampaign.3.vm")))
          .thenReturn("a third string");
      final ChatContext chatContext = new ChatContext();
      chatContext.setStep(2);
      chatContext.setAdditionalProperty("campaign_name", name);
      when(fakeTgChat.getChatContext()).thenReturn(chatContext);

      final Update thirdMessageDescription =
          gson.fromJson(
              "{\n"
                  + "\"update_id\":10000,\n"
                  + "\"message\":{\n"
                  + "  \"date\":1441645532,\n"
                  + "  \"chat\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"type\": \"private\",\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"message_id\":1365,\n"
                  + "  \"from\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"text\":\""
                  + description
                  + "\"\n"
                  + "}\n"
                  + "}",
              Update.class);

      final CompletableFuture<Optional<BaseRequest<?, ?>>> got =
          createCampaign.process(fakeTgChat, thirdMessageDescription);
      final Optional<BaseRequest<?, ?>> baseRequestOptional = got.get();
      final BaseRequest<?, ?> gotResponse = baseRequestOptional.get();
      assertEquals("a third string", gotResponse.getParameters().get("text"));

      verify(fakeCampaignUtil, times(1)).insertNewCampaign(name, description);
      verify(fakeChatUtil, times(1))
          .updateChatContext(fakeTgChat, "/createCampaign", 3, Collections.emptyMap());
    }
  }
}
